/*
// @author Eik List
// @last-modified 2015-09-01
// This is free and unencumbered software released into the public domain.
//
// Anyone is free to copy, modify, publish, use, compile, sell, or
// distribute this software, either in source code form or as a compiled
// binary, for any purpose, commercial or non-commercial, and by any
// means.
//
// In jurisdictions that recognize copyright laws, the author or authors
// of this software dedicate any and all copyright interest in the
// software to the public domain. We make this dedication for the benefit
// of the public at large and to the detriment of our heirs and
// successors. We intend this dedication to be an overt act of
// relinquishment in perpetuity of all present and future rights to this
// software under copyright law.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// For more information, please refer to <http://unlicense.org/>
*/
#ifdef DBG
    #include <stdio.h>
#endif
#include <string.h>
#include "poet.h"

// ---------------------------------------------------------------------

static const unsigned char POLYNOMIAL = 0xE1;
static const unsigned char MSB_MASK = 0x01;

// ---------------------------------------------------------------------

#define TOP_HASH     aesfour_encrypt(ctx->x, ctx->x, &(ctx->aes_axu))
#define BOTTOM_HASH  aesfour_encrypt(ctx->y, ctx->y, &(ctx->aes_axu))

// ---------------------------------------------------------------------


static void print_block(const char *label, const uint8_t *c)
{
#ifdef DBG
    printf("%s: ", label);
    
    for (size_t i = 0; i < BLOCKLEN; i++) {
        printf("%02X", c[i]);
    }

    printf("\n");
#endif    
}


// ---------------------------------------------------------------------

static int compare_blocks(const unsigned char* a, 
                          const unsigned char* b, 
                          const size_t num_bytes)
{
    unsigned char result = 0;
    
    for (size_t i = 0; i < num_bytes; i++) {
        result |= a[i] ^ b[i];
    }
    
    return result;
}

// ---------------------------------------------------------------------

static inline void xor_block(block c, const block a, const block b)
{
    for (size_t i = 0; i < BLOCKLEN; i++) {
        c[i] = a[i] ^ b[i];
    }
}

// ---------------------------------------------------------------------

static inline void to_array(unsigned char* dst, 
                            const uint64_t* src, 
                            const unsigned n)
{
    for (size_t i = 0; i < n; i++) {
        for (size_t j = 0; j < 8; ++j) {
            dst[i*8+j] = (unsigned char)((src[i] >> (8*j)) & 0xFF);
        }
    }
}

// ---------------------------------------------------------------------

static void encode_length(block s, const uint64_t len) 
{
    memset(s, 0x00, BLOCKLEN);
    to_array(s, &len, 1);
}

// ---------------------------------------------------------------------

static void shift_right(block h)
{
    for (size_t i = BLOCKLEN-1; i > 0; --i) {
        h[i] = (h[i] >> 1) | (h[i-1] << 7);
    }

    h[0] = h[0] >> 1;
}

// ---------------------------------------------------------------------

static void gf128_double(block h)
{
    const unsigned char msb = (h[BLOCKLEN-1] & MSB_MASK);
    shift_right(h);
    h[0] ^= msb * POLYNOMIAL;
}

// ---------------------------------------------------------------------

void keysetup_encrypt_only(poet_ctx_t *ctx, const unsigned char key[KEYLEN_BITS])
{
    block ctr;
    AES_KEY aes_enc;

    memset(ctx->tau, 0, BLOCKLEN);
    memset(ctr, 0, BLOCKLEN);

    //  Generate block cipher key 
    aes_expand_enc_key(key, KEYLEN_BITS, &aes_enc);
    aes_encrypt(ctr, ctx->k, &aes_enc);

    aes_expand_enc_key(ctx->k, KEYLEN_BITS, &(ctx->aes_enc));

    //  Generate header key 
    ctr[BLOCKLEN - 1] = 1; 
    aes_encrypt(ctr, ctx->l, &aes_enc);

    //  Generate e-AXU hash-function keys 
    ctr[BLOCKLEN - 1] = 2; 
    aes_encrypt(ctr, ctx->k_axu, &aes_enc);
    aes_expand_enc_key(ctx->k_axu, KEYLEN_BITS, &(ctx->aes_axu));
}

// ---------------------------------------------------------------------

void keysetup(poet_ctx_t *ctx, const unsigned char key[KEYLEN_BITS])
{
    keysetup_encrypt_only(ctx, key);
    aes_expand_dec_key(ctx->k, KEYLEN_BITS, &(ctx->aes_dec));
}

// ---------------------------------------------------------------------

static void encode_parameters(block s, 
                              const unsigned long long num_blocks_per_part, 
                              const unsigned long long intermediate_taglen) 
{
    memset(s, 0x00, BLOCKLEN);
    
    for (size_t j = 0; j < 8; ++j) {
        s[j] = (unsigned char)((num_blocks_per_part >> (8*j)) & 0xFF);
        s[8+j] = (unsigned char)((intermediate_taglen >> (8*j)) & 0xFF);
    }
}

// ---------------------------------------------------------------------

void process_header(poet_ctx_t *ctx,
                    const unsigned char *header,
                    unsigned long long header_len)
{
#ifdef DBG
    int i;
    printf("===Processing header\n");
#endif
    block mask;
    block in;
    block out;
    uint64_t offset = 0;

    ctx->mlen = 0;
    memset(ctx->tau, 0, BLOCKLEN);
    memcpy(mask, ctx->l, BLOCKLEN);

    // Process parameters
#if DBG    
    printf("  ==Parameter block==\n");
#endif    
    encode_parameters(in, NUM_BLOCKS_PER_PART, INTERMEDIATE_TAGLEN);

    print_block("(A) Input      :", in);
    print_block("(A) Mask       :", mask);
    
    xor_block(in, in, mask);

    print_block("(A) AES_Input  :", in);
    aes_encrypt(in, out, &(ctx->aes_enc));
    xor_block(ctx->tau, out, ctx->tau);
    
    print_block("(A) AES_Output :", out);
    print_block("(A) Tau        :", ctx->tau);
    
    gf128_double(mask);
#if DBG    
    printf("  ==Block n ==");
#endif    
    while (header_len >= BLOCKLEN) {
        printf("\n");
        print_block("(A) Input      :", header + offset);
        print_block("(A) Mask       :", mask);
        xor_block(in, header + offset, mask);
        print_block("(A) AES_Input  :", in);
        aes_encrypt(in, out, &(ctx->aes_enc));
        xor_block(ctx->tau, out, ctx->tau);

        print_block("(A) AES_Output :", out);
        print_block("(A) Tau        :", ctx->tau);
        offset += BLOCKLEN;
        header_len -= BLOCKLEN;

        gf128_double(mask);
    }

#if DBG    
    printf("  ==Final block==\n");
#endif
    //  Final block 
    memset(in, 0, BLOCKLEN);
    memcpy(in, header + offset, header_len);
    in[header_len] = 0x80;
    print_block("(A) Input      :", in);
    print_block("(A) Mask       :", mask);
    xor_block(in, mask, in);
    aes_encrypt(in, out, &(ctx->aes_enc));
    print_block("(A) AES_Output :", out);
    
    
    printf("\n");
    print_block("(A) Mask(tau)           :", ctx->tau);
    xor_block(ctx->tau, out, ctx->tau);
    print_block("(A) AES_Input  :", ctx->tau);
    aes_encrypt(ctx->tau, ctx->tau, &(ctx->aes_enc));
    print_block("(A) AES_Output(new tau) :", ctx->tau);

    memcpy(ctx->x, ctx->tau, BLOCKLEN);
    memcpy(ctx->y, ctx->tau, BLOCKLEN);
    ctx->y[BLOCKLEN - 1] ^= 1;
    
    print_block("X :", ctx->x);    
    print_block("Y :", ctx->y);        
}

// ---------------------------------------------------------------------

static void encrypt_block(poet_ctx_t *ctx, 
                          const block in, 
                          block out)
{
    block tmp;
    TOP_HASH;
    xor_block(ctx->x, in, ctx->x);
    print_block("  Top Hash ", ctx->x);
    aes_encrypt(ctx->x, tmp, &(ctx->aes_enc)); // in, out, key
    print_block("  Main Enc ", tmp);
    BOTTOM_HASH;
    print_block("  Bot Hash ", ctx->y);
    xor_block(out, tmp, ctx->y); // result, a, b
    print_block("  AES_Output   ", out);
    memcpy(ctx->y, tmp, BLOCKLEN);
    ctx->mlen += BLOCKLEN_BITS;
}

// ---------------------------------------------------------------------

void encrypt_final(poet_ctx_t *ctx,
                   const unsigned char *plaintext,
                   unsigned long long plen,
                   unsigned char *ciphertext,
                   unsigned long long* clen, 
                   unsigned char tag[TAGLEN])
{
#ifdef DBG
    int i;
    printf("===Processing data\n");
#endif
    uint64_t offset = 0;
    block s;
    block tmp;
    block tmp2;

    if (clen) {
        *clen = plen;
    }

    while (plen > BLOCKLEN) {
        encrypt_block(ctx, (plaintext + offset), (ciphertext + offset));
        plen -= BLOCKLEN;
        offset += BLOCKLEN;
    }

    // Encrypt the message length
#if DBG
    printf("  ==(M) Message length\n");
#endif     
    ctx->mlen += plen * 8;
    encode_length(s, ctx->mlen);
    print_block("(M) AES_Input  :", ctx->x);
    aes_encrypt(s, s, &(ctx->aes_enc));
    print_block("(M) AES_Output :", tmp);

    // Last message block must be padded if necessary
    memcpy(tmp, plaintext + offset, plen);
    memcpy(tmp + plen, ctx->tau, BLOCKLEN - plen);

    // Process last block + generate the tag
#if DBG
    printf("  ==(M) Last block\n");
#endif         
    TOP_HASH;

    xor_block(tmp, s, tmp);
    xor_block(ctx->x, tmp, ctx->x);

    print_block("(M) AES_Input  :", ctx->x);
    aes_encrypt(ctx->x, tmp, &(ctx->aes_enc));
    print_block("(M) AES_Output :", tmp);

    BOTTOM_HASH;
    
    xor_block(tmp2, tmp, ctx->y);
    memcpy(ctx->y, tmp, BLOCKLEN);
    xor_block(tmp, s, tmp2);

#ifdef DBG
    printf("  ==(M) Tag generation\n");
    printf("S\t: "); for(i=0;i<16;i++) printf("%02X",s[i]); printf("\n");
    printf("X\t: "); for(i=0;i<16;i++) printf("%02X",ctx->x[i]); printf("\n");
    printf("Y\t: "); for(i=0;i<16;i++) printf("%02X",ctx->y[i]); printf("\n");
#endif    
    
    // Perform tag splitting if needed
    memcpy(ciphertext + offset, tmp, plen);
    memcpy(tag, tmp + plen, BLOCKLEN - plen);

    // Generate tag
    TOP_HASH;
    xor_block(ctx->x, ctx->tau, ctx->x);
    aes_encrypt(ctx->x, tmp, &(ctx->aes_enc));

    BOTTOM_HASH;
    xor_block(tmp, ctx->y, tmp);
    xor_block(tmp, ctx->tau, tmp);
    
    memcpy(tag + (BLOCKLEN - plen), tmp, plen);
}

// ---------------------------------------------------------------------

static void decrypt_block(poet_ctx_t *ctx,
                          const block in,
                          block out)
{
    block tmp;
    BOTTOM_HASH;
    
    xor_block(ctx->y, in, ctx->y);
    aes_decrypt(ctx->y, tmp, &(ctx->aes_dec));

    TOP_HASH;
    
    xor_block(out, tmp, ctx->x);
    memcpy(ctx->x, tmp, BLOCKLEN);
    ctx->mlen += BLOCKLEN_BITS;
}

// ---------------------------------------------------------------------

int decrypt_final(poet_ctx_t *ctx,
                  const unsigned char *ciphertext,
                  unsigned long long clen,
                  const unsigned char tag[TAGLEN],
                  unsigned char *plaintext, 
                  unsigned long long* plen)
{
#ifdef DBG
    int i;
    printf("===Processing data\n");
#endif
    uint64_t offset = 0;
    block s;
    block tmp;
    block tmp2;
    int alpha;
    int beta;

    if (plen) {
        *plen = clen;
    }

    while (clen > BLOCKLEN) {
        decrypt_block(ctx, ciphertext + offset, plaintext + offset);
        clen -= BLOCKLEN;
        offset += BLOCKLEN;
    }

    // Encrypt the message length
    ctx->mlen += clen * 8;
    encode_length(s, ctx->mlen);
    aes_encrypt(s, s, &(ctx->aes_enc));

    // Pad the final ciphertext block if necessary
    memcpy(tmp, ciphertext + offset, clen);
    memcpy(tmp + clen, tag, BLOCKLEN - clen);

    // Process last block and generate the tag
    BOTTOM_HASH;
    xor_block(tmp, s, tmp);

    xor_block(ctx->y, tmp, ctx->y);
    aes_decrypt(ctx->y, tmp, &(ctx->aes_dec));

    TOP_HASH;
    xor_block(tmp2, tmp, ctx->x);
    xor_block(tmp2, s, tmp2);
    memcpy(ctx->x, tmp, BLOCKLEN);
    
#ifdef DBG
    printf("S\t: "); for(i=0;i<16;i++) printf("%02X",s[i]); printf("\n");
    printf("X\t: "); for(i=0;i<16;i++) printf("%02X",ctx->x[i]); printf("\n");
    printf("Y\t: "); for(i=0;i<16;i++) printf("%02X",ctx->y[i]); printf("\n");
#endif        

    // Perform tag splitting if needed
    memcpy(plaintext + offset, tmp2, clen);
    alpha = compare_blocks(tmp2 + clen, ctx->tau, BLOCKLEN - clen);

    // Generate tag
    TOP_HASH;
    xor_block(ctx->x, ctx->tau , ctx->x);
    aes_encrypt(ctx->x, tmp, &(ctx->aes_enc));

    BOTTOM_HASH;
    xor_block(tmp, ctx->y, tmp);
    xor_block(tmp, ctx->tau, tmp);
    
#ifdef DBG
    printf("tmp\t: "); for(i=0;i<16;i++) printf("%02X",tmp[i]); printf("\n");
    printf("tag\t: "); for(i=0;i<16;i++) printf("%02X",tag[i]); printf("\n");
#endif        

    beta = compare_blocks(tmp, tag + (BLOCKLEN - clen), clen);
    return alpha | beta;
}

