#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import binascii
import cffi
import math
import os
import random
import math


import sys
from enum import Enum
from .options import routines
from pkg_resources import get_distribution, DistributionNotFound


__all__ = ['gen_random', 'gen_dataset', 'gen_test_routine',
           'gen_single', 'print_header']

from .__init__ import __version__


ffi = cffi.FFI()
ffi.cdef('''
    int crypto_aead_encrypt(
        unsigned char *c,unsigned long long *clen,
        const unsigned char *m,unsigned long long mlen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *nsec,
        const unsigned char *npub,
        const unsigned char *k
    );
    int crypto_aead_decrypt(
        unsigned char *m,unsigned long long *mlen,
        unsigned char *nsec,
        const unsigned char *c,unsigned long long clen,
        const unsigned char *ad,unsigned long long adlen,
        const unsigned char *npub,
        const unsigned char *k
    );
    ''')

def print_header(opts):
    '''
    Print header file
    '''
    ignore_opts = (
        'lib_path', 'lib_file', 'dest', 'pdi_file', 'sdi_file', 'do_file',
        'verify_lib', 'routines', 'verbose', 'dbg', 'mode', 'human_readable') \
        + routines

    sorted_vars = [x for x in sorted(vars(opts)) if x not in ignore_opts]

    txt = "# This file was auto-generated by aeadtvgen v{}\n" \
        .format(__version__)
    txt += '#'*79 + '\n'
    txt += '# Parameter:\n'
    txt += '#\n'

    for opt in sorted_vars:
        value = getattr(opts, opt)
        if opt == 'io':
            opt = 'io (W,SW)'
        elif opt == 'len_format':
            opt = 'len_format (AD, DATA)'
        elif opt == 'block_size_ad':
            if value == None:
                continue
        txt += "# {:22} - {}\n".format(opt, value)
    txt += '#'*79 + '\n\n'


    for file_name in [opts.pdi_file, opts.sdi_file, opts.do_file]:
        file_path = os.path.join(opts.dest,file_name)
        with open(file_path, 'w') as f:
            f.write('#'*79 + '\n')
            f.write('# {}\n'.format(file_name))
            f.write(txt)

    if opts.human_readable:
        file_path = os.path.join(opts.dest,'test_vectors.txt')
        try:
            os.remove(file_path)
        except FileNotFoundError:
            pass


class Opcode(Enum):
    encrypt = 2
    decrypt = 3
    loadkey = 4
    actkey  = 7

class Status(Enum):
    success = 14
    failure = 15

class Segment(Enum):
    ad      = 1
    npub_ad = 2
    ad_npub = 3
    pt      = 4
    ct      = 5
    ct_tag  = 6
    tag     = 8
    len     = 10
    key     = 12
    npub    = 13
    nsec_pt = 14
    nsec_ct = 15

txt_opcode = {
    Opcode.encrypt: 'Authenticated Encryption',
    Opcode.decrypt: 'Authenticated Decryption',
    Opcode.loadkey: 'Load Key',
    Opcode.actkey:  'Activate Key',}

txt_status = {
    Status.success: 'Success',
    Status.failure: 'Failure',}

txt_segment = {
    Segment.ad     : 'Associated Data',
    Segment.npub   : 'Npub',
    Segment.npub_ad: 'Npub || Associated Data',
    Segment.ad_npub: 'Associated Data || Npub',
    Segment.pt     : 'Plaintext',
    Segment.ct     : 'Ciphertext',
    Segment.ct_tag : 'Ciphertext || Tag',
    Segment.tag    : 'Tag',
    Segment.len    : 'Length',
    Segment.key    : 'Key',
    Segment.nsec_pt: 'Nsec (plain)',
    Segment.nsec_ct: 'Nsec (cipher)',}

def lenbytes(hexstr):
    try:
        return int(math.ceil(len(hexstr)/2))
    except TypeError:
        return 0

def get_len(format, ad_len, pt_len):
    ''' Get segment length data hex format '''

    # Convert length to binary
    bin = '{:0{f[0]}b}{:0{f[1]}b}'.format(ad_len, pt_len, f=format)
    # Make sure data is byte multiple
    bin += '0'*int((-len(bin)) % 8)
    # Convert binary to hex string
    size = int(math.ceil((format[0]+format[1])/8))
    hex = '{:0{s}X}'.format(int(bin, 2), s=int(size*2))
    return hex

def get_msg_format(format, ofile, decrypt):
    ''' Create a msg format for pdi/sdi file '''

    msg_format = []
    tag = []
    for sgt in format:
        # Segment type format conversion
        if (not ofile): # Input file
            if sgt == 'data':
                sgt = 'ct' if (decrypt) else 'pt'
            elif sgt == 'data_tag':
                sgt = 'ct_tag' if (decrypt) else 'pt'
            elif sgt == 'nsec':
                sgt = 'nsec_ct' if (decrypt) else 'nsec_pt'
            elif sgt == 'tag':
                if (not decrypt):
                    continue
        else:   # output file check
            if sgt in ('npub', 'npub_ad', 'ad_npub', 'ad', 'len'):
                continue
            elif sgt == 'data':
                sgt = 'ct' if (not decrypt) else 'pt'
            elif sgt == 'data_tag':
                sgt = 'ct_tag' if (not decrypt) else 'pt'
            elif sgt == 'nsec':
                sgt = 'nsec_ct' if (not decrypt) else 'nsec_pt'
            elif sgt == 'tag':
                if (decrypt):
                    continue
                else:   # Put tag in the last segment for output of encryption
                    tag.append(sgt)
                    continue
                    
        msg_format.append(sgt)
    return msg_format + tag

def get_test_vector_info(msgid, keyid, ad_len, pt_len, ct_len, decrypt):
    ''' Get a string of test vector information '''
    if (decrypt):
        optxt = txt_opcode[getattr(Opcode, 'decrypt')]
        data = 'Ct Size = {: 4}'.format(ct_len)
    else:
        optxt = txt_opcode[getattr(Opcode, 'encrypt')]
        data = 'Pt Size = {: 4}'.format(pt_len)
    txt  = '#### {}\n'.format(optxt)
    txt += '#### MsgID={: 3}, KeyID={: 3} '.format(msgid, keyid)
    txt += 'Ad Size = {: 4}, {}\n'.format(ad_len, data)
    return txt

def instr_info(f):
    def wrapper(*args, **kwargs):
        txt = '# Instruction: Opcode={}\n'.format(txt_opcode[args[1]])
        return '{}{}'.format(txt, f(*args, **kwargs))
    return wrapper

@instr_info
def build_instr(iowidth, opcode, msgid, keyid, ofile=0):
    ''' Generate instruction '''
    binstr = '{:04b}{:06b}{:06b}'.format(opcode.value, keyid, msgid)
    binstr += '0'*(-len(binstr) % iowidth)
    hexstr = '{:0{w}X}'.format(int(binstr, 2), w=int(iowidth/4))
    if (not ofile):
        return 'INS = {}\n'.format(hexstr)
    else:
        return '# INS:{}\n'.format(hexstr)


def sgmt_info(f):
    def wrapper(*args, **kwargs):
        # Argument dissection
        data  = args[0]
        sgt   = args[1]
        ofile = args[2]
        opts  = args[3]
        (iowidth, io_per_line) = args[4]
        (is_partial, is_eoi, is_eot, is_lst) = args[5]

        pt_txt = 'Partial={} '.format(is_partial) \
            if opts.add_partial and sgt in ['pt','ct','ct_tag'] else ''
        last_txt = 'EOI={} '.format(is_eoi) if (not ofile) else ''
        txtsgmt = txt_segment[getattr(Segment, sgt)]
        txt = '# Info : {:>24}, '.format(txtsgmt)
        txt += '{}{}EOT={}, Last={}, Length={} bytes\n'.format(
            pt_txt, last_txt, is_eot, is_lst, lenbytes(data))
        return '{}{}'.format(txt, f(*args, **kwargs))
    return wrapper

def sgmt_hdr(f):
    def wrapper(*args, **kwargs):
        # Argument dissection
        data  = args[0]
        sgt   = args[1]
        ofile = args[2]
        opts  = args[3]
        (iowidth, io_per_line) = args[4]
        (is_partial, is_eoi, is_eot, is_lst) = args[5]

        is_eoi = is_eoi if (not ofile) else 0
        is_partial = is_partial \
            if opts.add_partial and sgt in ['pt','ct','ct_tag'] else 0
        code = getattr(Segment, sgt).value
        binstr = '{:04b}{}{}{}{}'.format(code, is_partial, is_eoi,
                                        is_eot, is_lst)
        binstr += '0'*8 + '{:016b}'.format(lenbytes(data))
        binstr += '0'*(-len(binstr) % iowidth)
        hexstr = '{:0{w}X}'.format(int(binstr, 2), w=int(iowidth/4))
        return 'HDR = {}\n{}'.format(hexstr, f(*args, **kwargs))
    return wrapper

@sgmt_info
@sgmt_hdr
def build_sgmt(data, sgt, ofile, opts, io_info, flags):
    ''' Generate a segment '''
    (iowidth, io_per_line) = io_info
    if len(data) > 0:
        bytes_per_line = iowidth/8*io_per_line
        tot_line = int(math.ceil(lenbytes(data)/bytes_per_line))
        txt = ''

        for i in range(tot_line):
            begin = int(i*bytes_per_line*2)
            end = begin + int(bytes_per_line*2)
            if i == tot_line-1:
                d = data[begin:]
                d += '00'*(-lenbytes(d) % int(iowidth/8))
                txt += 'DAT = {}\n'.format(d)
            else:
                txt += 'DAT = {}\n'.format(data[begin:end])
        return txt
    else :
        return ''


def build_status(iowidth):
    ''' Generate build status '''
    binstr = '{:04b}'.format(Status.success.value)
    binstr += '0'*(-len(binstr) % iowidth)
    hexstr = '{:0{w}X}'.format(int(binstr, 2), w=int(iowidth/4))
    txt = '# Status: Success\n'
    txt += 'STT = {}\n'.format(hexstr)
    return txt

class TestVector(object):
    ''' TestVector class '''
    BUFFER = '00'*128

    def __init__(self, opts, msg_id, key_id,
                 new_key, op, key, npub, nsec_pt, ad, pt):
        self.lib = ffi.dlopen(opts.lib_file)
        self.opts = opts
        self.key_id = key_id
        self.msg_id = msg_id
        self.new_key = new_key
        self.decrypt = op
        # Input
        self.key     = key
        self.npub    = npub
        self.nsec_pt = nsec_pt
        self.ad      = ad
        self.pt      = pt
        # Output
        self.nsec_ct = ''
        self.ct = ''
        self.tag = ''

    def aead_encrypt(self):
        ''' Compute aead algorithm '''

        pt_len  = lenbytes(self.pt)
        buf_len = lenbytes(self.BUFFER)
        # Prepare input to C function (add buffer to prevent overflow)
        m       = ffi.new("char[]", binascii.unhexlify(self.pt + self.BUFFER))
        mlen    = ffi.cast("unsigned long long", pt_len)
        c       = ffi.new("char[]", binascii.unhexlify('00'*(pt_len+buf_len)))
        clen    = ffi.new("unsigned long long *", pt_len+buf_len)
        ad      = ffi.new("char[]", binascii.unhexlify(self.ad + self.BUFFER))
        adlen   = ffi.cast("unsigned long long", lenbytes(self.ad))
        nsec    = ffi.new("char[]", binascii.unhexlify(self.nsec_pt))
        npub    = ffi.new("char[]", binascii.unhexlify(self.npub))
        key     = ffi.new("char[]", binascii.unhexlify(self.key))
        ### ABI level, in-line call
        self.lib.crypto_aead_encrypt(c, clen, m, mlen, ad,
                                     adlen, nsec, npub, key)
        # Convert output to Hexadecimal
        output = "".join(binascii.hexlify(c[i])
                   .decode("utf-8").upper() for i in range(0, clen[0]))

        nsec_ct = "".join(binascii.hexlify(nsec[i])
                         .decode("utf-8").upper()
                         for i in range(0, self.opts.nsec_size))
        ct_len = int((clen[0]-self.opts.tag_size/8)*2)
        ct  = output[:ct_len]
        tag = output[ct_len:int(clen[0]*2)]
        return (nsec_ct, ct, tag)

    def aead_decrypt(self):
        ''' Compute aead algorithm '''

        ct_len  = lenbytes(self.ct) + int(self.opts.tag_size/8)

        buf_len = lenbytes(self.BUFFER)
        # Prepare input to C function (add buffer to prevent overflow)
        m       = ffi.new("char[]", binascii.unhexlify('00'*ct_len))
        mlen    = ffi.new("unsigned long long *", ct_len)
        nsec    = ffi.new("char[]", binascii.unhexlify('00'*ct_len))
        c       = ffi.new("char[]", binascii.unhexlify(self.ct + self.tag))
        clen    = ffi.cast("unsigned long long", ct_len)
        ad      = ffi.new("char[]", binascii.unhexlify(self.ad + self.BUFFER))
        adlen   = ffi.cast("unsigned long long", lenbytes(self.ad))
        npub    = ffi.new("char[]", binascii.unhexlify(self.npub))
        key     = ffi.new("char[]", binascii.unhexlify(self.key))
        ### ABI level, in-line call
        auth_result = self.lib.crypto_aead_decrypt(m, mlen, nsec,
                                     c, clen, ad, adlen, npub, key)
        # Convert output to Hexadecimal
        pt = "".join(binascii.hexlify(m[i])
               .decode("utf-8").upper() for i in range(0, mlen[0]))
        nsec_pt = "".join(binascii.hexlify(nsec[i])
                    .decode("utf-8").upper()
                    for i in range(0, self.opts.nsec_size))
        return (auth_result, nsec_pt, pt)

    def get_data(self, sgt):
        ''' Get data based on segment type '''

        if (sgt == 'npub_ad'):
            data = getattr(self, 'npub') + getattr(self, 'ad')
        elif (sgt == 'ad_npub'):
            data = getattr(self, 'ad') + getattr(self, 'npub')
        elif (sgt == 'ct_tag'):
            data = getattr(self, 'ct') + getattr(self, 'tag')
        elif (sgt == 'len'):
            len_ad = lenbytes(self.ad)
            if self.decrypt:
                len_data = lenbytes(self.ct)
            else:
                len_data = lenbytes(self.pt)
            data = get_len(self.opts.len_format,
                           len_ad, len_data)
        else:
            data = getattr(self, sgt)
        return data

    def is_last_vld_segment(self, sgt, msg_format):
        ''' Determine whether this segment is the last valid data segment '''
        ignore_sgts = ('len','tag')
        if sgt in ignore_sgts:
            return False
        if len(self.get_data(sgt)) <= 0:
            return False

        for i in range(msg_format.index(sgt)+1,len(msg_format)):
            if msg_format[i] in ignore_sgts:
                continue
            if len(self.get_data(msg_format[i])) > 0:
                return False
        return True

    def gen_tv(self):
        ''' Generate test vector files based on provided options '''

        (self.nsec_ct, self.ct, self.tag) = self.aead_encrypt()

        # Check for mismatching decrypted values and tag
        if (self.opts.verify_lib):
            (auth_result, nsec_pt, pt) = self.aead_decrypt()
            assert auth_result == 0
            assert nsec_pt == self.nsec_pt
            assert pt == self.pt

        (iow, iosw)  = self.opts.io
        io_info = (iow, self.opts.max_io_per_line)

        is_partial = 1 if lenbytes(self.pt) % int(self.opts.block_size/8) > 0 \
                       else 0

        # PDI and SDI file
        for ofile, file_name in enumerate([self.opts.pdi_file,
                                           self.opts.do_file]):
            # Open file
            file_path = os.path.join(self.opts.dest, file_name)
            f = open(file_path, 'a')

            # Write Header
            txt = get_test_vector_info(self.msg_id,
                                       self.key_id,
                                       lenbytes(self.ad),
                                       lenbytes(self.pt),
                                       lenbytes(self.ct),
                                       self.decrypt)
            f.write('{}'.format(txt))

            if (not ofile):
                # Write New key
                if self.new_key:
                    txt = build_instr(iow, Opcode.actkey,
                                      self.msg_id, self.key_id)
                    f.write('{}'.format(txt))

            # Instruction
            opcode = Opcode.decrypt if self.decrypt else Opcode.encrypt
            txt = build_instr(iow, opcode, self.msg_id, self.key_id, ofile)
            f.write('{}'.format(txt))

            # Write Segment
            msg_format = get_msg_format(self.opts.msg_format,
                                        ofile, self.decrypt)

            for i, sgt in enumerate(msg_format):
                # Data getter
                data = self.get_data(sgt)

                # Sub-segment
                max_sgmt = (self.opts.block_size/8)*self.opts.max_block_per_sgmt
                tot_sgmt = int(math.ceil(lenbytes(data)/max_sgmt))
                tot_sgmt = 1 if tot_sgmt == 0 else tot_sgmt

                (is_eoi, is_eot, is_lst) = (0,0,0)
                for j in range(tot_sgmt):
                    begin = int(j*max_sgmt*2)
                    end = begin + int(max_sgmt*2)
                    if j == tot_sgmt-1:
                        is_eoi = \
                            int(self.is_last_vld_segment(sgt, msg_format))
                        is_lst = 1 if i == len(msg_format)-1 else 0
                        is_eot = 1
                        d = data[begin:]
                    else:
                        d = data[begin:end]
                    txt = ''

                    # ::Special rule for ciphertext expansion::
                    # Separates the last block in its own segment for
                    # ciphertext expansion mode to accommodate possible
                    # expansion/truncation of data.
                    #
                    # For plaintext, the last block can be empty to acommodate
                    # expected padding value.
                    #
                    # For ciphertext, the last block cannot be empty.
                    if (j == tot_sgmt-1
                            and self.opts.ciph_exp
                            and sgt in ['pt','ct','ct_tag']):
                        # Split d into d[1], d[2] if condition applies
                        rem = lenbytes(d) % int(self.opts.block_size/8)
                        if (lenbytes(d) >= int(self.opts.block_size/8)):
                            if (rem > 0):
                                d = (d[:-int(rem*2)], d[-int(rem*2):])
                            else:
                                if (sgt == 'pt'):
                                    d = (d, '')
                                else:
                                    if (lenbytes(d) == \
                                            int(self.opts.block_size/8)):
                                        d = (d, )
                                    else:
                                        rem = int(self.opts.block_size/8)
                                        d = (d[:-int(rem*2)], d[-int(rem*2):])
                        else:
                            d = (d, )

                        # Add intermediate block if necessary
                        if len(d) > 1:
                            if len(d[1]) == 0:
                                eoi = is_eoi
                                is_eoi = 0
                            else:
                                eoi = 0
                            flags = (is_partial, eoi, 0, 0)
                            txt += build_sgmt(d[0], sgt, ofile,
                                              self.opts, io_info, flags)
                            d = d[1]
                        else:
                            d = d[0]
                    flags = (is_partial, is_eoi, is_eot, is_lst)
                    txt += build_sgmt(d, sgt, ofile,
                                      self.opts, io_info, flags)

                    f.write('{}'.format(txt))

            if (ofile):
                # Write success
                txt = build_status(iow)
                f.write('{}'.format(txt))

            f.write('\n')
            f.close()



        # ==========
        # SDI file
        # ==========
        if (not self.new_key):
            return

        io_info = (iosw, self.opts.max_io_per_line)
        flags = (0, 1, 1, 1)
        sgt = 'key'

        file_path = os.path.join(self.opts.dest, self.opts.sdi_file)
        f = open(file_path, 'a')
        # Instruction
        txt  = '#### MsgID={: 3}, KeyID={: 3}\n'.format(self.msg_id,
                                                        self.key_id)
        txt += build_instr(iosw, Opcode.loadkey, self.msg_id, self.key_id)
        # Segment
        data = self.get_data(sgt)
        txt += build_sgmt(data, sgt, 0,
                          self.opts, io_info, flags)
        f.write('{}\n'.format(txt))
        f.close()

    def gen_nist_tv(self):
        if not self.opts.human_readable:
            return
        file_path = os.path.join(self.opts.dest, 'test_vectors.txt')
        f = open(file_path, 'a')
        f.write("#### Msg {:>3}\n".format(self.msg_id))
        attrs = ['key', 'npub', 'nsec_pt', 'ad', 'pt', 'nsec_ct', 'ct', 'tag']
        for attr in attrs:
            if attr in ['nsec_pt', 'nsec_ct']:
                if (self.opts.nsec_size <= 0):
                    continue
            f.write("{:7} = {}\n".format(attr, getattr(self, attr)))
        f.write('\n')
        f.close()

# ======================
# Construct a data set
# ======================
def gen_data(bytes: int, mode=0, init='06') -> str:
    """ Generate random data """
    if (bytes == 0):
        return ''
    else:
        if (mode ==0):
            return '{d:0{s}X}'.format(s=bytes*2, d=random.randrange(256**bytes))
        else:
            init = int(init, 16)
            data = ''.join('{0:02X}'.format((j+init)%256) for j in range(bytes))
            return data



def gen_dataset(opts, routine, start_msg_no, start_key_no, mode=0):
    '''
    Generate random dataset based on the specified routine with the following
    format: [[NEW_KEY(Boolean), Encryption/Decryption(Boolean,
              AD_SIZE, DATA_SIZE],
              ...,
            ]
    '''
    dataset = []
    key_id = start_key_no-1

    # print(routine)
    for i, tv in enumerate(routine):
        new_key   = 1 if i == 0 else routine[i][0]
        decrypt = routine[i][1]
        if mode == 1:
            key  = '55'*int(opts.key_size/8)
            npub = 'B0'*int(opts.npub_size/8)
            nsec = '66'*int(opts.nsec_size/8)
            ad   = 'A0'*routine[i][2]
            data = 'FF'*routine[i][3]
        else:
            key  = gen_data(int(opts.key_size/8),   mode, '55')
            npub = gen_data(int(opts.npub_size/8),  mode, 'B0')
            nsec = gen_data(int(opts.nsec_size/8),  mode, '66')
            ad   = gen_data(routine[i][2],          mode, 'A0')
            data = gen_data(routine[i][3],          mode, 'FF')

        if new_key == 0:
            key = dataset[i-1].key
            #! Automatically use old value for decryption
            #! if the same key is used for the same ad and plaintext size
            if (decrypt and not dataset[i-1].decrypt
                and routine[i][2] == lenbytes(dataset[i-1].ad)
                and routine[i][3] == lenbytes(dataset[i-1].pt)):
                npub = dataset[i-1].npub
                nsec = dataset[i-1].nsec_pt
                ad   = dataset[i-1].ad
                data = dataset[i-1].pt

        key_id = key_id + new_key

        dataset.append(TestVector(opts, i+start_msg_no, key_id,
                                  new_key, decrypt,
                                  key, npub, nsec, ad, data))
    return dataset, i+start_msg_no, key_id

def gen_single(opts, start_msg_no, start_key_no):
    # print('gen_single')
    a = [1] + opts.gen_single
    return gen_dataset(opts, [a,], start_msg_no, start_key_no)

def gen_random(opts, start_msg_no, start_key_no):
    routine = []
    for i in range(opts.gen_random):
        new_key = random.randrange(2)
        operation = random.randrange(2)
        sizeAd  = random.randrange(opts.min_ad, opts.max_ad+1)
        sizeMsg = random.randrange(opts.min_d , opts.max_d +1)
        routine.append([new_key, operation, sizeAd, sizeMsg])
    return gen_dataset(opts, routine,
                       start_msg_no, start_key_no)


def gen_test_routine(opts, start_msg_no, start_key_no):
    # print('gen_test_routine')
    bsa = opts.block_size_ad if opts.block_size_ad != None else opts.block_size
    bsd = opts.block_size
    bsa, bsd = int(bsa/8), int(bsd/8)
    (start, stop, mode) = opts.gen_test_routine
    routine =  [[True ,     False,      0,         0    ],
                [False,     True,       0,         0    ],
                [True ,     False,      1,         0    ],
                [False,     True,       1,         0    ],
                [True ,     False,      0,         1    ],
                [False,     True,       0,         1    ],
                [True ,     False,      1,         1    ],
                [False,     True,       1,         1    ],
                [True ,     False,      bsa,       bsd  ],
                [False,     True,       bsa,       bsd  ],
                [True ,     False,      bsa-1,     bsd-1],
                [False,     True,       bsa-1,     bsd-1],
                [True ,     False,      bsa+1,     bsd+1],
                [False,     True,       bsa+1,     bsd+1],
                [True ,     False,      bsa*2,     bsd*2],
                [False,     True,       bsa*2,     bsd*2],
                [True ,     False,      bsa*3,     bsd*3],
                [False,     True,       bsa*3,     bsd*3],
                [True ,     False,      bsa*4,     bsd*4],
                [False,     True,       bsa*4,     bsd*4],
                [True ,     False,      bsa*5,     bsd*5],
                [False,     True,       bsa*5,     bsd*5]]
    return gen_dataset(opts, routine[start-1:stop],
                       start_msg_no, start_key_no, mode)
